"""
This program and the accompanying materials are made available under the terms of the
Eclipse Public License v2.0 which accompanies this distribution, and is available at
https://www.eclipse.org/legal/epl-v20.html
SPDX-License-Identifier: EPL-2.0

Copyright Contributors to the Zincware Project.

Description: init file for the utils package.
"""



"""
This program and the accompanying materials are made available under the terms of the
Eclipse Public License v2.0 which accompanies this distribution, and is available at
https://www.eclipse.org/legal/epl-v20.html
SPDX-License-Identifier: EPL-2.0

Copyright Contributors to the Zincware Project.

Description: Module for the computation of so3 data
"""
from symdet.data.data_generator import DataGenerator
from typing import Union
import numpy as np
import matplotlib.pyplot as plt


class SO3(DataGenerator):
    """
    Class for the double well potential implementation.

    Attributes
    ----------
    noise : bool
                If true, noise is included in the data generation.
    variance : float
            Variance to use in the noise generation.
    radius : float
            Radius if the circle.
    radial_values : Union[float, list]
            Radial values to use in the data generation.

    See Also
    --------
    symdet.data.data_generator.DataGenerator

    Examples
    --------
    >>> from symdet import DoubleWellPotential
    >>> generator = SO3()
    >>> generator.load_data()
    >>> generator.plot_data()
    """

    def __init__(self, noise: bool = True, variance: float = 0.05, radius: float = 1.0):
        """
        Constructor for the double well potential.

        Parameters
        ----------
        noise : bool
                If true, noise is included in the data generation.
        variance : float
                Variance to use in the noise generation.
        radius : float
                Radius if the circle.
        """
        super().__init__()
        self.noise = noise
        self.variance = variance
        self.radius = radius
        self.radial_values = None

    def _sphere(self, points: int):
        """
        Generate point along a double well potential range.

        Parameters
        ----------
        points: int
                Number of points to use.

        Returns
        -------

        """
        if self.noise:
            self.radial_values = np.random.uniform(self.radius - self.variance,
                                                   self.radius + self.variance,
                                                   points)
        else:
            self.radial_values = self.radius

        theta = np.random.rand(points) * (np.pi * 2)  # generate angles randomly
        phi = np.random.rand(points) * np.pi  # generate angles randomly

        x = self.radial_values * np.cos(theta) * np.sin(phi)
        y = self.radial_values * np.sin(theta) * np.sin(phi)
        z = self.radial_values * np.cos(phi)

        self.domain = np.array(list(zip(x, y, z)))

    def load_data(self, points: Union[int, np.ndarray], save: bool = False):
        """
        Load / generate the data.

        Parameters
        ----------
        points : Union[int, np.ndarray]
                Points to generate, either an np.ndarray or an integer. If an integer, N points will be generated, if
                an array, it will either be treated as input to a function to generate values or those indices will be
                drawn from a pool.
        save : bool
                If true, save the data after generating it.

        Returns
        -------
        Updates the class state.
        """
        # generate domain and image data.
        if type(points) is int:
            self._sphere(points)

        # set domain and generate image data.
        else:
            raise ValueError(f"Type {type(points)} is not valid for this data generator, try an integer")

    def plot_data(self, save: bool = False, show: bool = True):
        """
        Plot the data.

        Parameters
        ----------
        save : bool
                If true, save the plot.
        show : bool (default=True)
                If true, show the result
        Returns
        -------
        Plots the data.
        """
        if self.domain is None:
            self._sphere(points=100)

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.scatter(self.domain[:, 0],
                   self.domain[:, 1],
                   self.domain[:, 2],
                   marker=".",
                   color="k")
        if save:
            plt.savefig(f"SO(2)_{len(self.domain)}.svg",
                        dpi=800,
                        format="svg")
        plt.show()

    def build_clusters(self, **kwargs):
        """
        Split the raw function data into classes.

        Returns
        -------
        Updates the class state.

        Notes
        -----
        Not required for this data.
        """
        pass

    def plot_clusters(self, save: bool = False):
        """
        Plot the clusters generated.

        Parameters
        ----------
        save

        Notes
        -----
        Not required for this analysis.

        """
        pass



"""
This program and the accompanying materials are made available under the terms of the
Eclipse Public License v2.0 which accompanies this distribution, and is available at
https://www.eclipse.org/legal/epl-v20.html
SPDX-License-Identifier: EPL-2.0

Copyright Contributors to the Zincware Project.

Group Detection
===============
Cluster raw data into symmetry groups
"""
from symdet.ml_models.dense_model import DenseModel
from symdet.analysis.model_visualization import Visualizer
from typing import Tuple
import numpy as np
import tensorflow as tf
from symdet.utils.data_clustering import compute_com, compute_radius_of_gyration


class GroupDetection:
    """
    A class to cluster raw data into symmetry groups.

    Attributes
    ----------
    model : DenseModel
            Model to use in the group detection.
    data_clusters : dict
            Data cluster class used for the partitioning of the data.
    representation_set : str
            Which set to use in the representation, train, validation, or test.
    """

    def __init__(self, model: DenseModel, data_clusters: dict, representation_set: str = 'train'):
        """
        Constructor for the GroupDetection class.

        Parameters
        ----------
        model : DenseModel
                Model to use in the group detection.
        data_clusters : dict
                Data cluster class used for the partitioning of the data.
        representation_set : str
                Which set to use in the representation, train, validation, or test.
        """
        self.model = model
        self.data = data_clusters
        self.representation_set = representation_set
        self.model.add_data(self.data)  # add the data to the model.

    def _get_model_predictions(self) -> Tuple:
        """
        Train the attached model.

        Returns
        -------
        val_data : tf.Tensor
                Data on which the prediction were made.
        model_predictions : Tuple
                Embedding layer of the NN on validation data.
        """
        self.model.train_model()
        if self.representation_set == 'train':
            val_data = self.model.train_ds
            predictions = self.model.model.predict(val_data[:, 0:self.model.input_shape])
        elif self.representation_set == 'test:':
            val_data = self.model.test_ds
            predictions = self.model.model.predict(val_data[:, 0:self.model.input_shape])
        else:
            val_data = self.model.val_ds
            predictions = self.model.model.predict(val_data[:, 0:self.model.input_shape])

        return val_data, predictions

    def _run_visualization(self):
        """
        Perform a visualization on the TSNE data.

        Returns
        -------

        """
        pass

    @staticmethod
    def _cluster_detection(function_data: np.ndarray, data: np.ndarray):
        """
        Use the results of the TSNE reduction to extract clusters.

        Parameters
        ----------
        function_data : tf.Tensor
                A tensor of the raw data to be collected.
        data : np.ndarray
                Results of the tsne representation

        Returns
        -------
        clusters : dict
                An unordered point cloud of data belonging to the same cluster.
                e.g. {1: [radial values], 2: [radial_values], ...}
        """
        net_array = np.concatenate((data, function_data), 1)
        sorted_data = tf.gather(net_array, tf.argsort(net_array[:, -1])).numpy()
        class_array = np.unique(function_data[:, -1])

        point_cloud = {}
        # loop over the class array
        for i, item in enumerate(class_array):
            start = np.searchsorted(sorted_data[:, -1], item, side='left')
            stop = np.searchsorted(sorted_data[:, -1], item, side='right') - 1
            com = compute_com(sorted_data[start:stop, 0:2])
            rg = compute_radius_of_gyration(sorted_data[start:stop, 0:2], com)

            #print(f"Class: {item}, COM: {com}, Rg: {rg}")
            if rg > 1000:
                point_cloud[item] = sorted_data[start:stop, 2:-1]

        return point_cloud

    @staticmethod
    def _filter_data(predictions: tf.Tensor, targets: tf.Tensor):
        """
        Check which data points are predicted well and include them in the data.

        Parameters
        ----------
        targets : tf.Tensor
                Target values on which predictions were made.
        predictions : tf.Tensor
                Network predictions.

        Returns
        -------

        """
        accepted_candidates = np.zeros(len(predictions))
        target_values = tf.keras.utils.to_categorical(targets[:, -1])
        counter = 0
        for i, item in enumerate(predictions):
            if np.linalg.norm(predictions[i] - target_values[i]) <= 2e-1:
                accepted_candidates[counter] = i
                counter += 1
        accepted_candidates = tf.convert_to_tensor(accepted_candidates[0:counter], dtype=tf.int32)

        return tf.gather(targets, accepted_candidates)

    def run_symmetry_detection(self, plot: bool = True, save: bool = False):
        """
        Run the symmetry detection routine.

        Parameters
        ----------
        plot : bool
                Plot the TSNE visualization.
        save : bool
                Save the image plotted.
        Returns
        -------
        None
        """
        validation_data, predictions = self._get_model_predictions()
        accepted_data = self._filter_data(predictions, validation_data)
        representation = self.model.get_embedding_layer_representation(accepted_data)  # get the embedding layer

        visualizer = Visualizer(representation, accepted_data[:, -1])
        data = visualizer.tsne_visualization(plot=plot, save=save)

        # determine coupled groups in the tSNE representation.
        point_cloud = self._cluster_detection(validation_data, data)

        return point_cloud



"""
__init__ file for the symdet package
"""
import os
from symdet.data.double_well_potential import DoubleWellPotential
from symdet.ml_models.dense_model import DenseModel
from symdet.symmetry_group_extraction.group_detection import GroupDetection
from symdet.generator_extraction.generators import GeneratorExtraction
from symdet.data.so2_data import SO2
from symdet.data.so3_data import SO3

__all__ = ['DoubleWellPotential', 'DenseModel', 'GroupDetection', 'SO2', 'SO3']
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"



"""
This program and the accompanying materials are made available under the terms of the
Eclipse Public License v2.0 which accompanies this distribution, and is available at
https://www.eclipse.org/legal/epl-v20.html
SPDX-License-Identifier: EPL-2.0

Copyright Contributors to the Zincware Project.

Description: Example data generator for the double well potential.
"""
from symdet.data.data_generator import DataGenerator
from symdet.utils.data_clustering import range_binning
from typing import Union
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
from tqdm import tqdm


class DoubleWellPotential(DataGenerator):
    """
    Class for the double well potential implementation.

    Attributes
    ----------
    a : float
            Scaling factor in the equation (see below)

    See Also
    --------
    symdet.data.data_generator.DataGenerator

    Examples
    --------
    >>> from symdet import DoubleWellPotential
    >>> generator = DoubleWellPotential()
    >>> generator.load_data()
    >>> generator.plot_data()

    Notes
    -----
    The double well potential is written mathematically:

    .. math:: V = -a \cdot (x^{2} + y^{2}) + (x^{2} + y^{2})^{2}

    We will require the data to be stored as x,y  coordinates in order to facilitate the generator extraction in this
    part of the process.
    """

    def __init__(self, a: float = 2.3):
        """
        Constructor for the double well potential.

        Parameters
        ----------
        a : float
                Scaling factor of the double well potential.
        """
        super().__init__()
        self.a = a

    def _double_well(self):
        """
        Generate point along a double well potential range.
        Returns
        -------

        """
        square_radii = tf.reduce_sum(tf.math.square(self.domain), 1)
        self.image = -self.a * square_radii + tf.square(square_radii)

    def _pick_points(self, n_points: int, min_val: float = 0, max_val: float = 1.6):
        """
        Generate random coordinates in the double well domain.

        Parameters
        ----------
        n_points : int
                Number of points to pick.
        min_val : float
                Minimum value to consider.
        max_val : float
                Maximum value to consider.

        Returns
        -------

        """
        self.domain = tf.random.uniform(shape=(n_points, 2),
                                        minval=min_val,
                                        maxval=max_val)

    def load_data(self, points: Union[int, np.ndarray], save: bool = False):
        """
        Load / generate the data.

        Parameters
        ----------
        points : Union[int, np.ndarray]
                Points to generate, either an np.ndarray or an integer. If an integer, N points will be generated, if
                an array, it will either be treated as input to a function to generate values or those indices will be
                drawn from a pool.
        save : bool
                If true, save the data after generating it.

        Returns
        -------
        Updates the class state.
        """
        # generate domain and image data.
        if type(points) is int:
            self._pick_points(points)
            self._double_well()

        # set domain and generate image data.
        else:
            self.domain = tf.convert_to_tensor(points)
            self._double_well()

    def plot_clusters(self, save: bool = False):
        """
        Plot the clusters generated.

        Parameters
        ----------
        save

        Returns
        -------

        """
        self.plot_data(show=False)

        for i, item in tqdm(enumerate(self.clustered_data), ncols=70, total=len(self.clustered_data)):
            r = tf.norm(self.clustered_data[item]['domain'], axis=1)
            v = self.clustered_data[item]['image']
            plt.plot(r, v, '.', label=f"Class {i}", markersize=15)

        plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))

        plt.show()

    def plot_data(self, save: bool = False, show: bool = True):
        """
        Plot the data.

        Parameters
        ----------
        save : bool
                If true, save the plot.
        show : bool (default=True)
                If true, show the result
        Returns
        -------
        Plots the data.
        """
        if self.domain is None:
            self._pick_points(1000, min_val=0, max_val=1.2)
            self._double_well()
        plt.plot(tf.norm(self.domain, axis=1), self.image, '.')
        plt.xlabel('r')
        plt.ylabel('V')
        plt.xlim(-0.03, 1.7)
        plt.ylim(-1.5, 1.3)
        plt.grid()
        if save:
            plt.savefig(f'Double_Well_{len(self.domain)}.svg', dpi=600, format='dpi')
        if show:
            plt.show()

    def build_clusters(self, value_range: list = None, bin_operation: list = None, representatives=1000):
        """
        Split the raw function data into classes.

        Parameters
        ----------
        representatives : int
            Number of class representatives to have for each bin.
        value_range : list
            The parameters within which to bin e.g.  k in [-5, 5]
        bin_operation : list
            Operation to apply to the bins e.g [1/5, 1e-3] will lead
            to bins of the form [k/5 - 1e-3, k/5 + 1e-3]

        Returns
        -------
        Updates the class state.

        Notes
        -----
        In the double well potential we can simply use the range_binning clustering algorithm.
        """
        # Replace None type parameters.
        if bin_operation is None:
            bin_operation = [1 / 5, 1e-3]
        if value_range is None:
            value_range = [-5, 5]
        n_classes = (value_range[1] - value_range[0]) + 1

        # Generate data if it does not exist
        if self.domain is None:
            self.load_data(n_classes * representatives * 1000)
        if len(self.domain) < n_classes * representatives * 1000:
            print("Loading additional data.")
            self.load_data(n_classes * representatives * 1000)

        self.clustered_data = range_binning(image=self.image,
                                            domain=self.domain,
                                            value_range=value_range,
                                            bin_operation=bin_operation,
                                            representatives=representatives)



"""
This program and the accompanying materials are made available under the terms of the
Eclipse Public License v2.0 which accompanies this distribution, and is available at
https://www.eclipse.org/legal/epl-v20.html
SPDX-License-Identifier: EPL-2.0

Copyright Contributors to the Zincware Project.

Description: Methods to help with clustering data.
"""
import tensorflow as tf
import numpy as np
from typing import Tuple
import sys


def _build_condlist(data: np.array, bin_values: dict) -> Tuple:
    """
    Build the condition list for the piecewise implementation.

    Parameters
    ----------
    data : np.array
            Data on which to apply conditions.
    bin_values : np.array
            Bin numbers for the constraint.

    Returns
    -------
    conditions : list
            These are conditions applied to the data.
    classes : list
            Class keys.
    """

    conditions = []
    classes = []
    for key in bin_values:
        conditions.append(
            np.logical_and(
                data >= (bin_values[key][0]), data <= (bin_values[key][1])
            )
        )
        classes.append(key)

    return conditions, classes


def _function_to_bins(function_values: tf.Tensor, bin_values: dict) -> tf.Tensor:
    """
    Sort function values into bins.

    Parameters
    ----------
    function_values : np.array
            Function values corresponding to radii values.
    bin_values : dict
            bin dictionary where keys are the class numbers.

    Returns
    -------
    conditions : tf.Tensor
            Conditions from the cond list build.
    """

    conditions, functions = _build_condlist(function_values, bin_values)

    return tf.convert_to_tensor(conditions)


def range_binning(
        image: tf.Tensor,
        domain: tf.Tensor,
        value_range: list,
        bin_operation: list,
        representatives: int = 100) -> dict:
    """
    A method to apply simple range binning to some data.

    Parameters
    ----------
    image : tf.Tensor
            data to cluster.
    domain : tf.Tensor
            data pool to return clustered.
    representatives : int
            Number of class representatives to have for each bin.
    value_range : list
            The parameters within which to bin e.g.  k in [-5, 5]
    bin_operation : list
            Operation to apply to the bins e.g [1/5, 1e-3] will
            lead to bins of the form [k/5 - 1e-3, k/5 + 1e-3]

    Returns
    -------
    classes : dict
            Data class numbers and their data representatives as a dictionary.
    """
    # Construct the classes and their range.
    bin_values = {}
    n_classes = (value_range[1] - value_range[0]) + 1

    for k in np.linspace(value_range[0], value_range[1], n_classes, dtype=int):
        bin_values[f"{k + abs(value_range[0])}"] = [
            bin_operation[0] * k - bin_operation[1],
            bin_operation[0] * k + bin_operation[1],
        ]

    # Collect the bin masks
    bin_masks = _function_to_bins(image, bin_values)

    # Check that there is enough data in each class.
    bin_count = tf.reduce_sum(tf.cast(bin_masks, tf.int8), 1)
    if any(bin_count) < representatives:
        print("WARNING: Not enough data! Some classes will be under-represented.")

    class_keys = list(bin_values.keys())

    clustered_data = {}
    for i in range(len(class_keys)):
        clustered_data[class_keys[i]] = {}
        filtered_domain = tf.boolean_mask(domain, bin_masks[i])
        filtered_image = tf.boolean_mask(image, bin_masks[i])
        clustered_data[class_keys[i]]['domain'] = filtered_domain[0:representatives]
        clustered_data[class_keys[i]]['image'] = filtered_image[0:representatives]

    return clustered_data


def compute_com(data: np.ndarray):
    """
    Compute the center of mass of some data.

    Parameters
    ----------
    data : np.ndarray
            Data on which to compute the center of mass.

    Returns
    -------

    """
    return tf.reduce_mean(data, axis=0)


def compute_radius_of_gyration(data: np.ndarray, com: np.ndarray):
    """
    Compute the radius of gyration of some data.

    Parameters
    ----------
    data : np.ndarray
    com : np.ndarray

    Returns
    -------

    """
    rg_primitive = tf.reduce_sum((data - com)**2, axis=1)

    return tf.reduce_mean(rg_primitive, axis=0)



"""
This program and the accompanying materials are made available under the terms of the
Eclipse Public License v2.0 which accompanies this distribution, and is available at
https://www.eclipse.org/legal/epl-v20.html
SPDX-License-Identifier: EPL-2.0

Copyright Contributors to the Zincware Project.

Generators
==========
Python module to extract generators from data.
"""
import tensorflow as tf
import numpy as np
from tqdm import tqdm
import random
from sklearn.linear_model import LinearRegression
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
from typing import Tuple


class GeneratorExtraction:
    """
    Class to extract generators from point clouds

    Attributes
    ----------
    point_cloud : tf.Tensor
            Point cloud on which to perform regression.
    delta : float
                Width of the hyperplane to be considered.
    epsilon : float
                Distance between points to be considered as a point pair.
    candidate_runs : int
                Number of times to generate candidates before running PCA. Tune if convergence is not found.
    basis : tf.Tensor
            Orthonormal basis of the point cloud.
    hyperplane_set : tf.Tensor
            Set of all points in the hyperplane.
    point_pairs : list
            A list of tuples where each tuple value is an index of a point. The tuple indices correspond to point pairs.
            i.e. if the tuple is (0, 400), then indices 0 and 400 in the hyperplane set are a pair connected by the
            application/s of the generators.
    dimension : int
            Dimensionality of the points.
    generator_candidates : list
            Generator candidates on which to perform PCA.
    constrained_generators : np.ndarray:
            Constrained generator candidates.
    """

    def __init__(
        self,
        point_cloud: tf.Tensor,
        delta: float = 0.5,
        epsilon: float = 0.3,
        candidate_runs: int = 10,
    ):
        """
        Constructor for the GeneratorExtraction class.

        Parameters
        ----------
        point_cloud : tf.Tensor
                Point cloud on which to perform regression
        delta : float
                Width of the hyperplane to be considered
        epsilon : float
                Distance between points to be considered as a point pair.
        candidate_runs : int
                Number of times to generate candidates before running PCA.
                 Tune if convergence is not found.
        """
        self.point_cloud = point_cloud
        self.delta = delta
        self.epsilon = epsilon
        self.candidate_runs = candidate_runs

        self.basis: tf.Tensor
        self.hyperplane_set: tf.Tensor
        self.point_pairs: list
        self.dimension = self._get_dimension()

        self.generator_candidates = []
        self.constrained_generators = []

    def _get_dimension(self):
        """
        Get the dimension of the points in the point cloud.

        Returns
        -------
        dimension : int
                Dimension of the points in the point cloud.
        """
        return len(self.point_cloud[0])

    def _remove_redundancy(self):
        """
        Remove the redundancy in the data with PCA.

        Returns
        -------
        Updates the class state.
        Notes
        -----
        Currently not implemented as it is not required.
        """
        pass

    def _generate_basis_set(self, gs_precision: int):
        """
        Build the basis set.

        Parameters
        ----------
        gs_precision : int
                Number of decimals after 0 to consider orthogonal.

        Returns
        -------
        Updates the basis set and enforces a check.
        """
        basis = list(self._start_gs())

        if self.dimension > 2:
            basis_candidates = np.zeros((self.dimension, self.dimension))
            for i, vector in enumerate(basis_candidates):
                vector[i] = 1
            reduced_candidates = self._eliminate_closest_vector(
                basis, list(basis_candidates)
            )
            for item in reduced_candidates:
                basis.append(self._perform_gs(item, basis))

        self.basis = tf.convert_to_tensor(basis)  # set the class attribute
        self._gs_check(gs_precision)

    def _gs_check(self, gs_precision: int):
        """
        Check to see that all basis vectors are orthogonal to one another.

        If this assert fails, the session will end.

        Parameters
        ----------
        gs_precision : int
                Number of decimals after 0 to consider orthogonal.

        Returns
        -------
        Will throw an exception if the assert fails.
        """
        for basis in self.basis:
            np.testing.assert_almost_equal(np.linalg.norm(basis), 1)  # check the normalization.
            for test in self.basis:
                if all(test == basis):
                    continue
                np.testing.assert_almost_equal(np.dot(basis, test), 0, decimal=gs_precision)  # check the orthogonality.

    def _perform_gs(self, vector: list, basis_set: list) -> np.ndarray:
        """
        Perform the Gram-Schmidt procedure.

        Parameters
        ----------
        vector : list
                Vector to orthonormalize
        basis_set : list
                Current basis set
        Returns
        -------
        basis_vector : list
                Orthonormalized basis vector
        """
        basis_vector = vector
        for basis_item in basis_set:
            basis_vector -= self._projection_operator(basis_item, basis_vector)

        return basis_vector / np.linalg.norm(basis_vector)

    def _eliminate_closest_vector(
            self,
            reference_vectors: list,
            test_vectors: list) -> np.ndarray:
        """
        Remove the closest vectors in the theoretical basis set

        Parameters
        ----------
        reference_vectors : list
                First two basis vectors
        test_vectors : list
                Vectors to test against
        Returns
        -------
        basis_vectors : np.ndarray
                Returns the basis vectors which minimizes the sum of
                squares of the scalar product.
        """
        distances = []
        for vector in test_vectors:
            d_1 = np.dot(vector, reference_vectors[0])
            d_2 = np.dot(vector, reference_vectors[1])
            distances.append(d_1 ** 2 + d_2 ** 2)

        return np.array(test_vectors)[np.argsort(distances)][: int(self.dimension - 2)]

    def _start_gs(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Pick the two first random basis vectors.

        Returns
        -------
        basis_vectors : list
                first two randomly selected, normalized basis vectors.
        """
        index_1 = random.randint(0, len(self.point_cloud) - 1)
        index_2 = random.randint(0, len(self.point_cloud) - 1)

        vector_1 = self.point_cloud[index_1] / np.linalg.norm(self.point_cloud[index_1])
        vector_2 = self.point_cloud[index_2] - self._projection_operator(
            vector_1, self.point_cloud[index_2]
        )
        vector_2 /= np.linalg.norm(vector_2)

        return vector_1, vector_2

    @staticmethod
    def _projection_operator(u, v) -> np.ndarray:
        """
        Perform the projection of u onto v.

        Returns
        -------
        proj_{u}(v) : np.ndarray
                Returns the projection of u on v. This is required for the Gram-Schmidt process.
        """
        return (np.dot(u, v) / np.dot(u, u)) * u

    def _construct_hyperplane_set(self):
        """
        Build the hyperplane set.

        Returns
        -------
        Updates the class state
        """
        if self.dimension == 2:
            self.hyperplane_set = self.point_cloud
        else:
            self.hyperplane_set = []
            for point in self.point_cloud:
                truth_table = []
                for vector in self.basis[2:]:
                    truth_table.append(np.dot(point, vector) < self.delta)

                if all(truth_table):
                    self.hyperplane_set.append(point)
        self.hyperplane_set = tf.convert_to_tensor(self.hyperplane_set)

    def _identify_point_pairs(self):
        """
        Identify point pairs within the hyperplane set.

        Returns
        -------
        Updates the class state.
        """
        self.point_pairs = []
        for index, origin in enumerate(self.hyperplane_set):
            for pair_index, reference in enumerate(self.hyperplane_set):
                # Skip the same point
                if pair_index == index:
                    continue
                else:
                    distance = np.linalg.norm(origin - reference)
                    if distance < self.epsilon:
                        self.point_pairs.append((index, pair_index))

    def _perform_regression(self):
        """
        Perform regression on the point pairs and extract generator candidates.

        Returns
        -------
        Updates the class state.
        """
        self._simple_regression()

        if self.dimension > 2:
            self._constrain_generators()

    def _constrain_generators(self):
        """
        If the dimension is greater than 2, the generators should be constrained by Equation 15 in the referenced paper.

        Returns
        -------
        Updates the class.
        """
        for generator in self.generator_candidates:
            outcome = []
            for basis in self.basis[2:]:
                data = np.matmul(
                    generator.reshape(self.dimension, self.dimension), basis
                )
                outcome = np.concatenate((outcome, data))

            if np.allclose(outcome, [0 for _ in range(len(outcome))], atol=1):
                self.constrained_generators.append(generator)

    def _full_regression(self):
        """
        Perform full regression to extract the generators.

        Returns
        -------
        Updates the class state.
        """
        self._simple_regression()
        unconstrained_generators = np.array(self.generator_candidates).reshape(
            (len(self.generator_candidates), self.dimension, self.dimension)
        )
        self.generator_candidates = []
        for generator in unconstrained_generators:
            truth_array = []
            for vector in self.basis[2:]:
                test = sum(generator * vector)
                truth_array.append(all(abs(test) < 1))
            if all(truth_array):
                self.generator_candidates.append(
                    np.array(generator).reshape((1, self.dimension * self.dimension))
                )

    def _simple_regression(self):
        """
        In the case where additional constraints are not needed, we simply perform regression on the problem to
        extract generator candidates.

        Returns
        -------
        Updates the class state.
        """
        Y = []
        X = []
        for pair in self.point_pairs:
            points = [self.hyperplane_set[pair[0]], self.hyperplane_set[pair[1]]]
            sigma = self._compute_sigma(points)
            Y.append(
                ((points[0] - points[1]) * np.linalg.norm(points[0]))
                / (sigma * np.linalg.norm(points[1] - points[0]))
            )
            X.append(points[0])

        generator = []
        for i in range(self.dimension):
            generator = np.concatenate(
                (generator, LinearRegression().fit(X, np.array(Y)[:, i]).coef_)
            )

        self.generator_candidates.append(generator)

    def _compute_sigma(self, pair) -> int:
        """
        Compute the directional information about the point pair.

        Parameters
        ----------
        pair : list
                A point pair from the point cloud.
        Returns
        -------
        sigma : int
                A direction measurement used to identify direction information in the basis set.
        """
        return np.sign(
            (np.dot(pair[0], self.basis[0]) * np.dot(pair[1], self.basis[1]))
            - (np.dot(pair[0], self.basis[1]) * np.dot(pair[1], self.basis[0]))
        )

    def _extract_generators(self, pca_components: int, factor: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """
        Perform PCA on candidates and extract true generators.

        Parameters
        ----------
        pca_components : int
                Number of pca components to use.

        Returns
        -------
        pca_components : np.ndarray
                pca components.
        factor : bool
                If true, factor the values by the dimensionality.
        variance : np.ndarray
                The explained variance list.
        """
        if self.dimension > 2:
            pca = PCA(n_components=pca_components)
            pca.fit(self.constrained_generators)

        pca = PCA(n_components=pca_components)
        pca.fit(self.generator_candidates)
        if factor:
            return np.sqrt(self.dimension) * pca.components_, pca.explained_variance_ratio_
        else:
            return pca.components_, pca.explained_variance_ratio_

    def _plot_results(self, std_values: list, save: bool = False):
        """
        Plot the results of the analysis.

        Parameters
        ----------
        std_values : list
                explained variance list to be plotted.
        save : bool
                If true, the plot will be saved.

        Returns
        -------
        Plots and image and saves it if required.
        """
        plt.plot([i for i in range(len(std_values))], std_values, "o-")
        plt.xlabel("No. PCA Components")
        plt.ylabel("Explained Variance (%)")
        plt.xticks(np.arange(len(std_values)), [i + 1 for i in range(len(std_values))])
        plt.xlim(-0.1, len(std_values) - 0.9)
        plt.ylim(-0.1, 1.1)
        plt.grid()
        if save:
            plt.savefig("PCA_STD.svg", dpi=800, format="svg")
        plt.show()

    def perform_generator_extraction(
            self,
            pca_components: int = 4,
            plot: bool = False,
            save: bool = False,
            factor: bool = True,
            gs_precision: int = 5) -> Tuple:
        """
        Collect all methods and perform the generator extraction.

        Parameters
        ----------
        pca_components : int
                Number of pca components to checked in the reduction.
        plot : bool
                If True, the outcomes will be plotted.
        factor : bool
                If true, factor the values by the dimensionality.
        save : bool
                If True, and plot is also True, the plots will be saved.
        gs_precision : int
                Number of decimals after 0 to consider orthogonal.

        Returns
        -------
        generators : list
                Return a list of generators.
        std_array : list
                explained variance list.
        """

        for _ in tqdm(range(self.candidate_runs),
                      ncols=100,
                      desc="Producing generator candidates"):
            try:
                self._remove_redundancy()
                self._generate_basis_set(gs_precision)
                self._construct_hyperplane_set()
                self._identify_point_pairs()
                self._perform_regression()
            except ValueError:
                continue

        generators, std_array = self._extract_generators(pca_components=pca_components,
                                                         factor=factor)
        for i, item in enumerate(generators):
            print(f"Principle Component {i + 1}: Explained Variance: {std_array[i]}")
            print(item.reshape((self.dimension, self.dimension)))
            print("\n")

        if plot:
            self._plot_results(std_array, save=save)

        return generators, std_array



"""
This program and the accompanying materials are made available under the terms of the
Eclipse Public License v2.0 which accompanies this distribution, and is available at
https://www.eclipse.org/legal/epl-v20.html
SPDX-License-Identifier: EPL-2.0

Copyright Contributors to the Zincware Project.

Model Visualization
===================
Visualize the NN models in different ways
"""
import matplotlib.pyplot as plt
from sklearn.manifold import TSNE
import numpy as np


class Visualizer:
    """
    Class for the visualization of NN models

    Attributes
    ----------
    data : tf.Tensor
                data to be visualized by the TSNE
        colour_map : list
                A colour map to be applied to the data so that it matches that of the cluster map.
    """

    def __init__(self, data, colour_map):
        """
        Constructor the visualizer class

        Parameters
        ----------
        data : tf.Tensor
                data to be visualized by the TSNE
        colour_map : list
                A colour map to be applied to the data so that it matches that of the cluster map.
        """

        self.data = data
        self.colour_map = colour_map

    def tsne_visualization(self,
                           perplexity=50,
                           n_components=2,
                           plot: bool = True,
                           save: bool = False) -> np.ndarray:
        """
        Display a TSNE representation of the models embedding layer

        Parameters
        ----------
        perplexity : int
                Perplexity of the tsne representation.
        n_components : int
                Dimensionality of the TSNE representation.
        plot : bool
                If true the TSNE representation will be plotted.
        save : bool
                If true the plot will be saved.

        Returns
        -------
        tsne_representation : np.ndarray
                Returns the tsne representation as an array.
        Notes
        -----
        See the theory documentation for a full overview of these parameters, particularly in the case of the TSNE
        values.
        """
        tsne_model = TSNE(n_components=n_components,
                          perplexity=perplexity,
                          random_state=1)
        tsne_representation = tsne_model.fit_transform(self.data)

        if plot:
            plt.scatter(
                tsne_representation[:, 0],
                tsne_representation[:, 1],
                c=self.colour_map,
                marker='.',
                cmap="viridis",
                vmax=11,
                vmin=-1,
            )
            plt.colorbar()
            if save:
                plt.savefig(
                    f"tsne_representation_{perplexity}_{n_components}.svg",
                    dpi=800,
                    format="svg",
                )
            plt.show()

        return tsne_representation



"""
This program and the accompanying materials are made available under the terms of the
Eclipse Public License v2.0 which accompanies this distribution, and is available at
https://www.eclipse.org/legal/epl-v20.html
SPDX-License-Identifier: EPL-2.0

Copyright Contributors to the Zincware Project.

Description: Parent class for the data generator. Data should be passed to the group detection and generator extraction
routines through a data generator child class.
"""
import abc
from typing import Union
import numpy as np
import tensorflow as tf


class DataGenerator(metaclass=abc.ABCMeta):
    """
    A class to generate data for use in the Symmetry analysis.

    Attributes
    ----------
    domain : tf.Tensor
            Domain values of the function.
    image : tf.Tensor
            Image values of the function, i.e. f(x) for all x.
    image_size : int
            Size of the data pool.
    domain_shape : tuple
            Shape of the domain points.
    clustered_data : dict
            A dictionary of clustered data.
    """

    def __init__(self):
        """
        Constructor for the DataGenerator class.
        """
        self.domain = None
        self.image = None
        self.image_size = None
        self.domain_shape = None
        self.clustered_data = None

    def plot_data(self, save: bool = False):
        """
        Plot the data.

        Parameters
        ----------
        save : bool
                If true the figure will be saved.

        Returns
        -------

        """
        raise NotImplementedError

    def plot_clusters(self, save: bool = False):
        """
        Plot the data clusters.

        Parameters
        ----------
        save : bool
                If true the figure will be saved.

        Returns
        -------

        """
        raise NotImplementedError

    def load_data(self, points: Union[int, np.ndarray], save: bool = False):
        """
        Load some data either from a computation or from a pool into the class state.

        Parameters
        ----------
        points : Union[int, np.ndarray]
                Points to generate, either an np.ndarray or an integer. If an integer, N points will be generated, if
                an array, it will either be treated as input to a function to generate values or those indices will be
                drawn from a pool.
        save : bool
                If true, save the data after generating it.

        Returns
        -------

        """
        raise NotImplementedError

    def build_clusters(self, **kwargs):
        """
        Split the raw function data into classes.

        Parameters
        ----------
        **kwargs

        Returns
        -------
        Updates the class state.

        Notes
        -----
        In the double well potential we can simply use the range_binning clustering algorithm.
        """
        raise NotImplementedError



"""
This program and the accompanying materials are made available under the terms of the
Eclipse Public License v2.0 which accompanies this distribution, and is available at
https://www.eclipse.org/legal/epl-v20.html
SPDX-License-Identifier: EPL-2.0

Copyright Contributors to the Zincware Project.

Dense Model
===========
Dense neural network model
"""
import tensorflow as tf
import numpy as np
from tensorflow.keras import regularizers
from tensorflow.keras.layers import Input


class DenseModel:
    """
    Class for the construction and training of a dense NN model

    Attributes
    ----------
    data_dict : dict
            Dictionary of data where the key is the class name and the values are the coordinates belongin to that
            class. This is fundamentall a classification problem!
    n_layers : int
            Number of hidden layers to use.
    units : int
            Number of units to use in each layer.
    epochs : int
            Number of epochs during training.
    activation : str
            Activation function to use.
    monitor : str
            Monitor parameter for use in training.
    lr : float
            Learning rate of the algorithm.
    batch_size : int
            batch size for the training.
    terminate_patience : int
            Patience value for termination to be used in a callback.
    lr_patience : int
            Patience in the learning rate reduction.
    train_ds : tf.data.Dataset
            Train dataset.
    test_ds : tf.data.Dataset
            Test dataset.
    val_ds : tf.data.Dataset
            validation dataset.
    model : tf.keras.Model
            Tensorflow model to train.
    """

    def __init__(
        self,
        n_layers: int = 5,
        units: int = 10,
        epochs: int = 20,
        activation: str = "relu",
        monitor: str = "accuracy",
        lr: float = 1e-4,
        batch_size: int = 100,
        terminate_patience: int = 10,
        lr_patience: int = 5):
        """
        Constructor fpr the Dense model class.

        Parameters
        ----------
        data_dict : dict
                Dictionary of data where the key is the class name and the values are the coordinates belonging to that
                class. This is fundamentally a classification problem!
        n_layers : int
                Number of hidden layers to use.
        units : int
                Number of units to use in each layer.
        epochs : int
                Number of epochs during training.
        activation : str
                Activation function to use.
        monitor : str
                Monitor parameter for use in training.
        lr : float
                Learning rate of the algorithm.
        batch_size : int
                batch size for the training.
        terminate_patience : int
                Patience value for termination to be used in a callback.
        lr_patience : int
                Patience in the learning rate reduction.
        """
        self.n_layers = n_layers
        self.data_dict = None
        self.input_shape = None
        self.units = units
        self.epochs = epochs
        self.activation = activation
        self.monitor = monitor
        self.batch_size = batch_size
        self.lr = lr
        self.terminate_patience = terminate_patience
        self.lr_patience = lr_patience
        self.train_ds = None
        self.test_ds = None
        self.val_ds = None
        self.model = None

    def add_data(self, data: dict):
        """
        Add cluster data to the model.

        Parameters
        ----------
        data : dict
                Cluster data to be added to the class.

        Returns
        -------

        """
        self.data_dict = data
        self.input_shape = len(self.data_dict[list(self.data_dict.keys())[0]]['domain'][0])

    def _shuffle_and_split_data(self):
        """
        shuffle and split the parsed dataset

        Returns
        -------
        Updates the class state.
        """

        for key in self.data_dict:
            labels = tf.repeat(
                tf.convert_to_tensor(np.array(key), dtype=tf.float32),
                len(self.data_dict[key]['domain']),
            )

            stacked_data = tf.concat([tf.cast(self.data_dict[key]['domain'], dtype=tf.float32),
                                      tf.transpose([labels])],
                                     axis=1)

            data_volume = len(stacked_data)
            train, test, validate = tf.split(
                stacked_data,
                [
                    int(0.5 * data_volume),
                    int(0.3 * data_volume),
                    int(0.2 * data_volume),
                ],
                axis=0,
            )
            if self.train_ds is None:
                self.train_ds = train
            else:
                self.train_ds = tf.concat([self.train_ds, train],
                                          axis=0)
            if self.test_ds is None:
                self.test_ds = test
            else:
                self.test_ds = tf.concat([self.test_ds, test],
                                         axis=0)
            if self.val_ds is None:
                self.val_ds = validate
            else:
                self.val_ds = tf.concat([self.val_ds, validate],
                                        axis=0)

        self.train_ds = tf.random.shuffle(self.train_ds)
        self.test_ds = tf.random.shuffle(self.test_ds)
        self.val_ds = tf.random.shuffle(self.val_ds)

    def _build_model(self):
        """
        Build the ml model

        Returns
        -------
        Updates the class state.
        """

        model = tf.keras.Sequential()
        input_layer = Input(shape=[self.input_shape])
        model.add(input_layer)

        # Loop over the layers excluding one for the input_layer, the second last, the embedding, and the softmax layer
        for i in range(self.n_layers - 2):
            model.add(
                tf.keras.layers.Dense(
                    self.units,
                    self.activation,
                    kernel_regularizer=regularizers.l2(0.0000),
                )
            )

        # Add the second-to-last layer
        model.add(
            tf.keras.layers.Dense(
                self.units,
                self.activation,
                name="second_last_layer",
                kernel_regularizer=regularizers.l2(0.0000),
            )
        )
        # Add the embedding layer
        model.add(
            tf.keras.layers.Dense(
                self.units,
                self.activation,
                name="embedding_layer",
                kernel_regularizer=regularizers.l2(0.001),
            )
        )

        # Add the softmax layer
        model.add(
            tf.keras.layers.Dense(
                len(self.data_dict), name="softmax_layer", activation="softmax"
            )
        )

        self.model = model  # Add the model to the class

    def _compile_model(self) -> list:
        """
        Compile the model

        Returns
        -------
        callback_list : list
                A list of model callbacks to be applied during training.
        """

        # Define the callbacks
        terminating_callback = tf.keras.callbacks.EarlyStopping(
            monitor=self.monitor,
            patience=self.terminate_patience,
        )
        reduction_callback = tf.keras.callbacks.ReduceLROnPlateau(
            monitor=self.monitor,
            patience=self.lr_patience,
            factor=0.1,
            mode="auto",
            min_lr=0.00001,
            cooldown=0,
        )

        opt = tf.keras.optimizers.Adam(learning_rate=self.lr, decay=0.0)

        self.model.compile(
            optimizer=opt, loss="categorical_crossentropy", metrics=["acc"]
        )

        return [terminating_callback, reduction_callback]

    def train_model(self):
        """
        Collect other methods and train the ML model
        """

        self._shuffle_and_split_data()  # Build the datasets
        self._build_model()  # Build the NN model
        self._compile_model()
        print(self.model.summary())  # Print a model summary

        # Train the model
        for i in range(1, 6):
            self.model.fit(
                x=self.train_ds[:, 0:self.input_shape],
                y=tf.keras.utils.to_categorical(self.train_ds[:, -1]),
                batch_size=self.batch_size,
                shuffle=True,
                validation_data=(
                    self.test_ds[:, 0:self.input_shape],
                    tf.keras.utils.to_categorical(self.test_ds[:, -1]),
                ),
                verbose=1,
                epochs=self.epochs,
            )

    def _evaluate_model(self):
        """
        Evaluate the tensorflow model on the validation data

        Returns
        -------
        Prints the model evaluation.
        """

        attributes = self.model.evaluate(
            x=self.val_ds[:, 0:self.input_shape], y=tf.keras.utils.to_categorical(self.val_ds[:, -1])
        )
        print(f"Loss: {attributes[0]} \n" f"Accuracy: {attributes[1]}")

    def get_embedding_layer_representation(self, data_array: np.ndarray) -> tf.Tensor:
        """
        Return the representation constructed by the embedding layer

        Parameters
        ---------
        data_array : np.array
                Data on which the model should be applied

        Returns
        -------
        predictions : tf.Tensor
                Predictions on the data_array returned in their high dimensional representation.
        """

        model = tf.keras.Sequential()
        input_data = Input(shape=[self.input_shape])
        model.add(input_data)
        for layer in self.model.layers[:-1]:
            model.add(layer)

        model.build()
        return model.predict(data_array[:, 0:self.input_shape])



"""
Setup.py file for the SymDet package.
"""

import setuptools
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, "requirements.txt")) as requirements_file:
    # Parse requirements.txt, ignoring any commented-out lines.
    requirements = [
        line
        for line in requirements_file.read().splitlines()
        if not line.startswith("#")
    ]

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="SymDet",
    version="0.0.1",
    author="Samuel Tovey",
    author_email="tovey.samuel@gmail.com",
    description="Symmetry detection and Lie generator extraction.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SamTov/SymDet",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
    python_requires=">=3.6",
    install_requires=requirements,
)
